@@ -13,7 +13,7 @@ with a few cherry-picks from the current implementation
 (<https://github.com/tjguk/active_directory/blob/master/active_directory.py>)
 
 Rewrite for Python3 with minimized dependencies
-by Rainer Schwarzbach, 2021-01-28
+by Rainer Schwarzbach, 2021-02-01
 
 License: MIT
 
@@ -34,10 +34,6 @@ import win32security
 #
 
 
-BASE_TIME = datetime.datetime(1601, 1, 1)
-TIME_NEVER_HIGH_PART = 0x7fffffff
-TIME_NEVER_KEYWORD = '<never>'
-
 ADO_COMMAND = 'ADODB.Command'
 ADO_CONNECTION = 'ADODB.Connection'
 CONNECTION_PROVIDER = 'ADsDSOObject'
@@ -74,8 +70,8 @@ def connection():
 def signed_to_unsigned(number):
     """Convert a signed integer to an unsigned one,
     adapted from the current upstream implementation
-    <https://github.com/tjguk/active_directory/
-     blob/master/active_directory.py>
+    <https://github.com
+     /tjguk/active_directory/blob/master/active_directory.py>
     """
     if number >= 0:
         return number
@@ -83,68 +79,89 @@ def signed_to_unsigned(number):
     return struct.unpack('L', struct.pack('l', number))[0]
 
 
-def convert_to_datetime(ad_time):
-    """Return a datetime from active directory.
-
-    numeric_date is the number of 100-nanosecond intervals
-    since 12:00 AM January 1, 1601, see
-    <https://ldapwiki.com/wiki/LargeInteger#section-LargeInteger-NumericDate>
-
-    Return 'never' for dates with a high part of 0x7fffffff.
-    If the time still exceeds the python datetime range,
-    return the maximum supported datetime.
-    """
-    if ad_time is None:
-        return None
-    #
-    high_part, low_part = [signed_to_unsigned(part) for part in
-                           (ad_time.HighPart, ad_time.LowPart)]
-    if high_part == TIME_NEVER_HIGH_PART:
-        return TIME_NEVER_KEYWORD
-    #
-    numeric_date = (high_part << 32) + low_part
-    delta = datetime.timedelta(microseconds=numeric_date / 10)
-    try:
-        return BASE_TIME + delta
-    except OverflowError:
-        return datetime.datetime.max
-    #
-
-
-def convert_to_guid(item):
-    """Return a GUID from an LDAP entry's property"""
-    if item is None:
-        return None
-    #
-    guid = convert_to_hex(item)
-    slice_borders = (8, 12, 16, 20)
-    return '{%s}' % '-'.join(
-        guid[slice(*pair)]
-        for pair in zip((None, *slice_borders), (*slice_borders, None)))
-
-
-def convert_to_hex(item):
-    """Return a hexadecimal representation of binary data"""
-    if item is None:
-        return None
-    #
-    return ''.join('%02x' % (char & 0xff) for char in bytes(item))
-
-
-def convert_to_sid(item):
-    """Return a PySID from binary data"""
-    if item is None:
-        return None
-    #
-    return win32security.SID(bytes(item))
-
-
 #
 # Classes
 #
 
 
-class UnsignedIntegerMapping():
+class Convert:
+
+    """A collection of converter methods for LdapEntry properties"""
+
+    base_time = datetime.datetime(1601, 1, 1)
+    time_never_high_part = 0x7fffffff
+    time_never_keyword = '<never>'
+
+    @classmethod
+    def to_datetime(cls, ad_time):
+        """Return a datetime from active directory.
+
+        numeric_date is the number of 100-nanosecond intervals
+        since 12:00 AM January 1, 1601, see
+        <https://ldapwiki.com
+         /wiki/LargeInteger#section-LargeInteger-NumericDate>
+
+        Return '<never>' for dates with a high part of 0x7fffffff.
+        If the time still exceeds the python datetime range,
+        return the maximum supported datetime.
+        """
+        if ad_time is None:
+            return None
+        #
+        high_part, low_part = [signed_to_unsigned(part) for part in
+                               (ad_time.HighPart, ad_time.LowPart)]
+        if high_part == cls.time_never_high_part:
+            return cls.time_never_keyword
+        #
+        numeric_date = (high_part << 32) + low_part
+        delta = datetime.timedelta(microseconds=numeric_date / 10)
+        try:
+            return cls.base_time + delta
+        except OverflowError:
+            return datetime.datetime.max
+        #
+
+    @classmethod
+    def to_guid(cls, item):
+        """Return a GUID from an LDAP entry's property"""
+        if item is None:
+            return None
+        #
+        guid = cls.to_hex(item)
+        slice_borders = (8, 12, 16, 20)
+        return '{%s}' % '-'.join(
+            guid[slice(*pair)]
+            for pair in zip((None, *slice_borders), (*slice_borders, None)))
+
+    @staticmethod
+    def to_hex(item):
+        """Return a hexadecimal representation of binary data"""
+        if item is None:
+            return None
+        #
+        return ''.join('%02x' % (char & 0xff) for char in bytes(item))
+
+    @staticmethod
+    def to_sid(item):
+        """Return a PySID from binary data"""
+        if item is None:
+            return None
+        #
+        return win32security.SID(bytes(item))
+
+    @staticmethod
+    def to_tuple(item):
+        """Return a tuple from the item"""
+        if not item:
+            return tuple()
+        #
+        if isinstance(item, str):
+            return (item,)
+        #
+        return tuple(item)
+
+
+class UnsignedIntegerMapping:
 
     """Mapping of names to unsigned integer numbers
     supporting lookups in each direction
@@ -265,11 +282,11 @@ USER_ACCOUNT_CONTROL = FlagsMapping(
     ADS_UF_TRUSTED_TO_AUTHENTICATE_FOR_DELEGATION=0x01000000)
 
 
-class RecordSet():
+class RecordSet:
 
     """Simple wrapper around an ADO Recordset, see
-    <https://docs.microsoft.com/windows/win32/adsi
-     /searching-with-activex-data-objects-ado>
+    <https://docs.microsoft.com
+     /windows/win32/adsi/searching-with-activex-data-objects-ado>
     """
 
     search_properties = dict(
@@ -491,14 +508,14 @@ class SearchFilter:
         self.__primary_key_name = primary_key_name
         self.__fixed_parameters = fixed_parameters
 
-    def execute_query(self, ldap_path, *args, **kwargs):
+    def execute_query(self, ldap_url, *args, **kwargs):
         """Build an SQL statement and execute a query
-        from the provided LDAP path.
+        starting at the provided LDAP url.
         Yield RecordSet objects.
         """
         sql_statement = '\n'.join([
             'SELECT ADsPath, userAccountControl',
-            'FROM %r' % ldap_path.url,
+            'FROM %r' % ldap_url,
             self.where_clause(*args, **kwargs)])
         for result in RecordSet.query(sql_statement):
             yield result
@@ -556,20 +573,45 @@ class LdapEntry:
     factory function.
     """
 
-    additional_conversions = {}
-    property_adspath = 'ADsPath'
-    property_ntsecuritydescriptor = 'nTSecurityDescriptor'
-    property_guid = 'GUID'
-    property_parent = 'Parent'
-    ignore_properties = {property_ntsecuritydescriptor}
-    additional_properties = {
-        property_adspath, property_guid, property_parent}
+    additional_properties = {'ADsPath'}
+    ignore_properties = {'nTSecurityDescriptor'}
+    conversions = dict(
+        accountExpires=Convert.to_datetime,
+        ADsPath=LdapPath.from_string,
+        badPasswordTime=Convert.to_datetime,
+        creationTime=Convert.to_datetime,
+        dSASignature=Convert.to_hex,
+        forceLogoff=Convert.to_datetime,
+        groupType=GROUP_TYPES.get_flag_names,
+        lastLogoff=Convert.to_datetime,
+        lastLogon=Convert.to_datetime,
+        lastLogonTimestamp=Convert.to_datetime,
+        lockoutDuration=Convert.to_datetime,
+        lockoutObservationWindow=Convert.to_datetime,
+        lockoutTime=Convert.to_datetime,
+        maxPwdAge=Convert.to_datetime,
+        member=Convert.to_tuple,
+        memberOf=Convert.to_tuple,
+        minPwdAge=Convert.to_datetime,
+        modifiedCountAtLastProm=Convert.to_datetime,
+        modifiedCount=Convert.to_datetime,
+        msExchMailboxGuid=Convert.to_guid,
+        objectGUID=Convert.to_guid,
+        objectSid=Convert.to_sid,
+        pwdLastSet=Convert.to_datetime,
+        replicationSignature=Convert.to_hex,
+        sAMAccountType=SAM_ACCOUNT_TYPES.get_name,
+        userAccountControl=USER_ACCOUNT_CONTROL.get_flag_names,
+        uSNChanged=Convert.to_datetime,
+        uSNCreated=Convert.to_datetime)
 
     def __init__(self, com_object):
         """Store properties form the provided COM object.
         The property names are determined from the schema,
         plus the required (cls.)additional_properties,
         minus the preformance-degrading (cls.)ignore_properties.
+        Additionally, properties that still are COM objects itself
+        or memoryviews after the conversion will be ignored.
         """
         schema = win32com.client.GetObject(com_object.Schema)
         property_names = tuple(
@@ -578,12 +620,6 @@ class LdapEntry:
             | set(schema.OptionalProperties)
             | self.additional_properties
             if single_property not in self.ignore_properties)
-        conversions = dict(
-            objectGUID=convert_to_guid,
-            uSNChanged=convert_to_datetime,
-            uSNCreated=convert_to_datetime,
-            replicationSignature=convert_to_hex)
-        conversions.update(self.additional_conversions)
         self.__case_translation = dict()
         self.__property_cache = dict()
         self.__empty_properties = set()
@@ -595,15 +631,15 @@ class LdapEntry:
                 continue
             #
             self.__case_translation[name.lower()] = name
+            try:
+                com_property = self.conversions[name](com_property)
+            except KeyError:
+                pass
+            #
             if com_property is None:
                 self.__empty_properties.add(name)
                 continue
             #
-            try:
-                com_property = conversions[name](com_property)
-            except KeyError:
-                pass
-            #
             self.__add_property(name, com_property)
         #
         self.items = self.__property_cache.items
@@ -614,18 +650,13 @@ class LdapEntry:
         return sorted(self.__empty_properties)
 
     @property
-    def parent(self):
-        """Return this object's parent LDAP entry"""
-        return produce_entry(self[self.property_parent])
-
-    @property
-    def path(self):
-        """Return the COM object's ADsPath"""
-        return LdapPath.from_string(self[self.property_adspath])
+    def ldap_url(self):
+        """Return the LDAP URL of the entry"""
+        return self.ADsPath.url
 
     def __add_property(self, name, value):
         """Add a property value only if it is not a
-        COM_Object or a memoryview (or a collection of those)
+        COM Object or a memoryview (or a collection of those)
         """
         if isinstance(value, (list, tuple)):
             if value and isinstance(
@@ -638,13 +669,14 @@ class LdapEntry:
         #
 
     def child(self, single_path_cmponent):
-        """Return the relative child of this entry. The relative_path
-        is inserted into this entry's LDAP path to make a coherent
-        LDAP path for a child entry, eg:
-
-        users = root.child('cn=Users')
+        """Return the relative child of this entry.
+        The relative_path must be a single LDAP path component.
+        It is prepended to this entry's LDAP path
+        to make a coherent LDAP path for a child entry.
         """
-        return produce_entry(LdapPath(single_path_cmponent, *self.path))
+        return produce_entry(
+            LdapPath(single_path_cmponent,
+                     *self.self.ADsPath.components))
 
     def print_dump(self):
         """Print all non-empty properties in
@@ -658,7 +690,7 @@ class LdapEntry:
 
     def __eq__(self, other):
         """Compare the GUIDs"""
-        return self[self.property_guid] == other[self.property_guid]
+        return self.objectGUID == other.objectGUID
 
     def __getattr__(self, name):
         """Instance attribute access to the com object's properties
@@ -688,38 +720,25 @@ class LdapEntry:
 
     def __hash__(self):
         """Identify by the GUID"""
-        return hash(self[self.property_guid])
+        return hash(self.objectGUID)
 
     def __repr__(self):
         """Return a representation with the class name
         and the path
         """
-        return "<%s: %s>" % (self.__class__.__name__, str(self.path))
+        return "<%s: %s>" % (self.__class__.__name__, str(self.ADsPath))
 
     def __str__(self):
         """Return the path"""
-        return str(self.path)
+        return str(self.ADsPath)
 
 
 class User(LdapEntry):
 
-    """Active Directory user with an additional bool property
-    (account_disabled)
+    """Active Directory user with an additional
+    convenience bool property (account_disabled)
     """
 
-    additional_conversions = dict(
-        accountExpires=convert_to_datetime,
-        badPasswordTime=convert_to_datetime,
-        lastLogoff=convert_to_datetime,
-        lastLogon=convert_to_datetime,
-        lastLogonTimestamp=convert_to_datetime,
-        lockoutTime=convert_to_datetime,
-        msExchMailboxGuid=convert_to_guid,
-        objectSid=convert_to_sid,
-        pwdLastSet=convert_to_datetime,
-        sAMAccountType=SAM_ACCOUNT_TYPES.get_name,
-        userAccountControl=USER_ACCOUNT_CONTROL.get_flag_names)
-
     @property
     def account_disabled(self):
         """Return True if the account is disabled"""
@@ -730,22 +749,13 @@ class Group(LdapEntry):
 
     """Active Directory group"""
 
-    additional_conversions = dict(
-        groupType=GROUP_TYPES.get_flag_names,
-        objectSid=convert_to_sid,
-        sAMAccountType=SAM_ACCOUNT_TYPES.get_name)
-
     def walk(self):
         """Yield a tuple of (self, subgroups_list, users_list)
         and repeat that (recursively) for each subgroup.
         """
-        member_paths = self.member or []
-        if isinstance(member_paths, str):
-            member_paths = [member_paths]
-        #
         groups_list = []
         users_list = []
-        for single_path in member_paths:
+        for single_path in self.member:
             child_entry = produce_entry(single_path)
             if isinstance(child_entry, self.__class__):
                 groups_list.append(child_entry)
@@ -761,134 +771,6 @@ class Group(LdapEntry):
         #
 
 
-class Computer(LdapEntry):
-
-    """Active Directory computer"""
-
-    additional_conversions = dict(
-        accountExpires=convert_to_datetime,
-        badPasswordTime=convert_to_datetime,
-        lastLogoff=convert_to_datetime,
-        lastLogon=convert_to_datetime,
-        lastLogonTimestamp=convert_to_datetime,
-        objectSid=convert_to_sid,
-        pwdLastSet=convert_to_datetime,
-        sAMAccountType=SAM_ACCOUNT_TYPES.get_name,
-        userAccountControl=USER_ACCOUNT_CONTROL.get_flag_names)
-
-
-class OrganizationalUnit(LdapEntry):
-
-    """Active Directory Organisational unit"""
-
-    user_search_fields = ('sAMAccountName', 'displayName', 'cn')
-
-    def find(self, *args, **kwargs):
-        """Return an LdapEntry for the first matching
-        search result.
-        """
-        for found_path in self.search(*args, **kwargs):
-            return produce_entry(found_path)
-        #
-        return None
-
-    def find_user(self, *args, **kwargs):
-        """Return a User object for the first matching
-        search result.
-        """
-        args_list = list(args)
-        try:
-            name = args_list.pop(0)
-        except IndexError:
-            pass
-        else:
-            user_search = []
-            for field_name in self.user_search_fields:
-                if field_name not in kwargs:
-                    user_search.append('%s=%r' % (field_name, str(name)))
-                #
-            #
-            if user_search:
-                args_list = [' OR '.join(user_search)] + args_list
-            #
-        #
-        for found_path in self.search(
-                *args_list, search_filter=SEARCH_FILTERS['userid'], **kwargs):
-            return produce_entry(found_path)
-        #
-
-    def search(self, *args, active=None, search_filter=None, **kwargs):
-        """Yield LDAP paths (plain strings) for all found Entries.
-
-        If 'active' is set to True or False explicitly,
-        yield the path only if the userAccountControl
-        property value matches the desired state.
-
-        if 'search_filter' is not set, determine a search filter
-        automatically.
-        """
-        if not isinstance(search_filter, SearchFilter):
-            for (keyword, candidate) in SEARCH_FILTERS.items():
-                try:
-                    value = kwargs.pop(keyword)
-                except KeyError:
-                    continue
-                #
-                kwargs['_primary_key_'] = value
-                search_filter = candidate
-                break
-            else:
-                search_filter = SearchFilter(None)
-            #
-        #
-        if active is None:
-            for result in search_filter.execute_query(
-                    self.path, *args, **kwargs):
-                yield result.ADsPath
-            #
-            return
-        #
-        bitmask = USER_ACCOUNT_CONTROL['ADS_UF_ACCOUNTDISABLE']
-        desired_state = bitmask
-        if active:
-            desired_state = 0
-        #
-        for result in search_filter.execute_query(self.path, *args, **kwargs):
-            try:
-                if result.userAccountControl & bitmask == desired_state:
-                    yield result.ADsPath
-                #
-            except TypeError:
-                # no userAccountControl property
-                yield result.ADsPath
-            #
-        #
-
-
-class DomainDNS(OrganizationalUnit):
-
-    """Active Directory Domain DNS"""
-
-    additional_conversions = dict(
-        creationTime=convert_to_datetime,
-        dSASignature=convert_to_hex,
-        forceLogoff=convert_to_datetime,
-        lockoutDuration=convert_to_datetime,
-        lockoutObservationWindow=convert_to_datetime,
-        maxPwdAge=convert_to_datetime,
-        minPwdAge=convert_to_datetime,
-        modifiedCount=convert_to_datetime,
-        modifiedCountAtLastProm=convert_to_datetime,
-        objectSid=convert_to_sid)
-
-
-class PublicFolder(LdapEntry):
-
-    """Active Directory public folder"""
-
-    ...
-
-
 #
 # Module-level functions
 #
@@ -915,17 +797,14 @@ def produce_entry(ldap_path, lazy=True):
         #
     #
     object_class_lower = com_object.Class.lower()
-    for ldap_entry_class in (
-            User, Group, DomainDNS,
-            OrganizationalUnit, Computer, PublicFolder):
-        if ldap_entry_class.__name__.lower() == object_class_lower:
+    for ldap_entry_subclass in (User, Group):
+        if ldap_entry_subclass.__name__.lower() == object_class_lower:
             return GLOBAL_CACHE.setdefault(ldap_path.url,
-                                           ldap_entry_class(com_object))
+                                           ldap_entry_subclass(com_object))
         #
     #
-    raise ValueError(
-        'Problem with object %s: No matching class %r found' % (
-            com_object, object_class_lower))
+    return GLOBAL_CACHE.setdefault(ldap_path.url,
+                                   LdapEntry(com_object))
 
 
 def root(server=None):
@@ -948,25 +827,101 @@ def root(server=None):
     #
 
 
-def find(*args, **kwargs):
-    """Find an LDAP entry.
-    Determine the type by the keyword argument.
-    Find a user using the keywords if no other entry type
-    could be determined.
+def search(*args,
+           active=None,
+           search_base=None,
+           search_filter=None,
+           **kwargs):
+    """Yield LDAP paths (plain strings) for all found entries.
+    Search starts at the LDAP URL specified in 'search_base'.
+    If that is not set, search from the  Active Directory root.
+
+    If 'active' is set to True or False explicitly,
+    yield the path only if the userAccountControl
+    property value matches the desired state.
+
+    if 'search_filter' is not set, determine a search filter
+    automatically.
     """
-    return root().find(*args, **kwargs)
+    if not isinstance(search_filter, SearchFilter):
+        for (keyword, candidate) in SEARCH_FILTERS.items():
+            try:
+                value = kwargs.pop(keyword)
+            except KeyError:
+                continue
+            #
+            kwargs['_primary_key_'] = value
+            search_filter = candidate
+            break
+        else:
+            search_filter = SearchFilter(None)
+        #
+    #
+    if not search_base:
+        search_base = root().ldap_url
+    #
+    query_results = search_filter.execute_query(search_base, *args, **kwargs)
+    if active is None:
+        for result in query_results:
+            yield result.ADsPath
+        #
+        return
+    #
+    bitmask = USER_ACCOUNT_CONTROL['ADS_UF_ACCOUNTDISABLE']
+    desired_state = bitmask
+    if active:
+        desired_state = 0
+    #
+    for result in query_results:
+        try:
+            if result.userAccountControl & bitmask == desired_state:
+                yield result.ADsPath
+            #
+        except TypeError:
+            # no userAccountControl property
+            yield result.ADsPath
+        #
+    #
 
 
-def find_user(*args, **kwargs):
+def search_users(*args, **kwargs):
+    """Yield LDAP paths (plain strings) for all found users."""
+    args_list = list(args)
+    try:
+        name = args_list.pop(0)
+    except IndexError:
+        pass
+    else:
+        user_search = []
+        for field_name in ('sAMAccountName', 'displayName', 'cn'):
+            if field_name not in kwargs:
+                user_search.append('%s=%r' % (field_name, str(name)))
+            #
+        #
+        if user_search:
+            args_list = [' OR '.join(user_search)] + args_list
+        #
+    #
+    kwargs['search_filter'] = SEARCH_FILTERS['userid']
+    return search(*args_list, **kwargs)
+
+
+def get_first_entry(*args, **kwargs):
+    """Return the LDAP entry for the first found match."""
+    for found_path in search(*args, **kwargs):
+        return produce_entry(found_path)
+    #
+    return None
+
+
+def get_first_user(*args, **kwargs):
     """Find a user by name or other properties
     from the cached root entry
     """
-    return root().find_user(*args, **kwargs)
-
-
-def search(*args, **kwargs):
-    """Search from the cached root entry"""
-    return root().search(*args, **kwargs)
+    for found_path in search_users(*args, **kwargs):
+        return produce_entry(found_path)
+    #
+    return None
 
 
 # vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:
